/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <limits.h>
#include <sys/wait.h>
#include "tcpjunk.h"
#include "utils.h"
#include "net.h"
#include "tags.h"

void server_main(struct tj_data *inst_data)
  {
  /* some vars */
  int r_timeout;
  int buf_id;
  int send_buf_len = inst_data->sess_data_len;
  char *send_buf = NULL;
  send_buf = tj_malloc(send_buf, send_buf_len);
  int keepalive_counter = inst_data->keepalive;

  /* SSL Stuff */
  int ssl_err;
  SSL_CTX *ctx = NULL;
  SSL *ssl = NULL;
  SSL_METHOD *meth;

  /* prepare SSL stuff if needed */
  if (inst_data->use_ssl == 1)
	{
	SSL_library_init();
	SSL_load_error_strings();

	/* SSL_METHOD structure */
	meth = SSLv3_method();
	ctx = SSL_CTX_new(meth);
	if ((ctx) == NULL)
	  {
	  printf("Cannot create an SSL_CTX structure\n");
	  goto finish;
	  }

	/* load server certificate */
	if (SSL_CTX_use_certificate_file(ctx, inst_data->server_cert,
		SSL_FILETYPE_PEM) <= 0)
	  {
	  sprintf(log_msg, "SSL certificate %s", ERR_error_string(ERR_get_error(),
		  NULL));
	  tj_log(inst_data, log_msg, c_red);
	  goto finish;
	  }

	/* load private-key */
	if (SSL_CTX_use_PrivateKey_file(ctx, inst_data->server_key,
		SSL_FILETYPE_PEM) <= 0)
	  {
	  sprintf(log_msg, "SSL key %s", ERR_error_string(ERR_get_error(), NULL));
	  tj_log(inst_data, log_msg, c_red);
	  goto finish;
	  }

	/* check certificate and key match */
	if (!SSL_CTX_check_private_key(ctx))
	  {
	  tj_log(inst_data, "SSL - Key does not match certificate", c_red);
	  goto finish;
	  }

	/* SSL structure */
	ssl = SSL_new(ctx);

	if ((ssl) == NULL)
	  {
	  tj_log(inst_data, "Cannot create SSL structure", c_red);
	  goto finish;
	  }

	/* Assign socket to SSL structure */
	SSL_set_fd(ssl, inst_data->srv_sock);

	/* SSL Handshake */
	ssl_err = SSL_accept(ssl);
	if ((ssl_err) == -1)
	  {
	  sprintf(log_msg, "SSL %s", ERR_error_string(ERR_get_error(), NULL));
	  tj_log(inst_data, log_msg, c_red);
	  goto finish;
	  }
	}

  /* choose session file automatically */
  if (inst_data->auto_file > 0)
	{
	set_auto_session_file(inst_data);

	/* load session file */
	load_session_data(inst_data);

	/* check session data */
	find_sections(inst_data);
	}

  /* start transfer */
  send_data:

  buf_id = 0;
  r_timeout = 0;

  /* get next send buffer from file */
  while (buf_id < inst_data->sect_idx)
	{
	/* see what client has to say on start */
	if (buf_id == 0 && inst_data->wait_conn == 1)
	  r_timeout = read_reply_tcp(inst_data->srv_sock, ssl, inst_data);

	if (r_timeout > 0)
	  {
	  tj_log(inst_data, "Session timeout", c_red);
	  goto finish;
	  }

	/* prepare a temp send buffer */
	if (send_buf_len != inst_data->sects_len[buf_id])
	  {
	  send_buf_len = inst_data->sects_len[buf_id];
	  send_buf = tj_realloc(send_buf, send_buf_len);
	  }

	memcpy(send_buf, inst_data->sects[buf_id], send_buf_len);

	/* process tags */
	send_buf_len = repeat_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = instance_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = count_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = char_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = fuzz_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = system_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = encode_tag(&send_buf, inst_data, send_buf_len);
	send_buf_len = charr_tag(&send_buf, inst_data, send_buf_len);

	/* send data to client */
	send_data_tcp(inst_data->srv_sock, ssl, inst_data, send_buf, send_buf_len);

	/* check what client has to say */
	if (!(inst_data->no_wait_last == 1 && buf_id == inst_data->sect_idx - 1))
	  {
	  r_timeout = read_reply_tcp(inst_data->srv_sock, ssl, inst_data);

	  if (r_timeout > 0)
		{
		tj_log(inst_data, "Session timeout", c_red);
		goto finish;
		}
	  }

	buf_id++;
	}

  /* if keep alive send again */
  if (keepalive_counter > 1 && r_timeout == 0)
	{
	keepalive_counter--;
	goto send_data;
	}

  /* complete */
  finish:

  free(send_buf);
  disconnect_tcp(inst_data->srv_sock, inst_data, ssl, ctx);

  if (inst_data->gui == 0)
	exit(EXIT_SUCCESS);
  else
	return;
  }

void server_init(struct tj_data *inst_data)
  {
  /* start server mode */
  int yes = 1;
  int rslt;
  int srv_sock;
  char bind_ip[128];

  struct sockaddr_storage cl_addr;
  socklen_t cl_addr_l;
  cl_addr_l = sizeof(cl_addr);

  /* turn off debug mode */
  if (inst_data->debug != 2)
	inst_data->debug = 0;

  /* set server bind address */
  if (inst_data->ipv6)
	strcpy(bind_ip, "0::0");
  else
	strcpy(bind_ip, "0.0.0.0");
  inst_data->local_addr = check_address(inst_data, bind_ip, inst_data->port,
	  ai_passive);

  /* allocate listening socket */
  srv_sock = socket(inst_data->local_addr->ai_family, SOCK_STREAM, 0);
  if (srv_sock < 0)
	printf("tcpjunk: Failed to allocate socket: %s\n", strerror(errno));

  /* set SO_REUSEADDR on socket */
  rslt = setsockopt(srv_sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int));
  if (rslt < 0)
	printf("tcpjunk: Error setting SO_REUSEADDR on socket: %s\n", strerror(
		errno));

  /* bind socket */
  rslt = bind(srv_sock, inst_data->local_addr->ai_addr,
	  inst_data->local_addr->ai_addrlen);
  if (rslt < 0)
	{
	printf("tcpjunk: Failed to bind socket: %s\n", strerror(errno));
	exit(EXIT_FAILURE);
	}

  /* listen on socket */
  rslt = listen(srv_sock, BACKLOG);
  if (rslt < 0)
	{
	printf("tcpjunk: Failed to listen on socket: %s\n", strerror(errno));
	exit(EXIT_FAILURE);
	}
  printf("tcpjunk: Listening on port %s, use -v/-d for info/data\n",
	  inst_data->port);

  /* server loop */
  for (;;)
	{
	inst_data->srv_sock = accept(srv_sock, (struct sockaddr *) &cl_addr,
		&cl_addr_l);

	if (inst_data->srv_sock < 0)
	  {
	  printf("tcpjunk: Failed accept on socket: %s\n", strerror(errno));
	  continue;
	  }
	else if (inst_data->verbose == 1)
	  printf("tcpjunk: TCP connection from %s\n", addr_ntop(
		  (struct sockaddr *) &cl_addr, cl_addr_l));

	/* create new server instance */
	if (!fork())
	  {
	  /* new child code */
	  close(srv_sock);
	  srand((unsigned int) (time(NULL) ^ getpid()) + getppid());
	  server_main(inst_data);
	  exit(EXIT_SUCCESS);
	  }
	wait(NULL);

	/* back in parent code */
	close(inst_data->srv_sock);

	/* increment instance number */
	inst_data->inst_num++;
	if (inst_data->inst_num == INT_MAX)
	  inst_data->inst_num = 1;
	}
  }
